/*
 * Copyright 2001 Hans Reiser
 */

#include "includes.h"



/* compares description block with commit block. returns 0 if they differ, 1
   if they match */
int does_desc_match_commit (struct reiserfs_journal_desc *desc, 
			    struct reiserfs_journal_commit *commit) 
{
    return (get_jcommit_trans_id (commit) == get_jdesc_trans_id (desc) &&
	    get_jcommit_len (commit) == get_jdesc_len (desc));
}


/* d_bh is descriptor, return number of block where commit block of this
   transaction is to be */
static unsigned long commit_expected (reiserfs_filsys_t * fs, struct buffer_head * d_bh)
{
    unsigned long offset;
    struct reiserfs_journal_desc * desc;


    desc = (struct reiserfs_journal_desc *)d_bh->b_data;
    offset = d_bh->b_blocknr - get_sb_journal_1st_block (fs->s_ondisk_sb);
    return get_sb_journal_1st_block (fs->s_ondisk_sb) + 
	((offset + get_jdesc_len (desc) + 1) % get_sb_journal_size (fs->s_ondisk_sb));
}


/* d_bh contains journal descriptor, return number of block where descriptor
   block of next transaction should be */
static unsigned long next_desc_expected (reiserfs_filsys_t * fs, struct buffer_head * d_bh)
{
    unsigned long offset;
    struct reiserfs_journal_desc * desc;


    desc = (struct reiserfs_journal_desc *)d_bh->b_data;
    offset = d_bh->b_blocknr - get_sb_journal_1st_block (fs->s_ondisk_sb);
    return get_sb_journal_1st_block (fs->s_ondisk_sb) + 
	((offset + get_jdesc_len (desc) + 2) % get_sb_journal_size (fs->s_ondisk_sb));
}


/**/
int is_valid_transaction (reiserfs_filsys_t * fs, struct buffer_head * d_bh)
{
    struct buffer_head * c_bh;
    struct reiserfs_journal_desc * desc;
    struct reiserfs_journal_commit * commit;
    

    if (!d_bh || who_is_this (d_bh->b_data, d_bh->b_size) != THE_JDESC)
	return 0;

    /* read expected commit block and compare with descriptor block */
    c_bh = bread (fs->s_journal_dev, commit_expected (fs, d_bh), fs->s_blocksize);
    if (!c_bh)
	return 0;

    desc = (struct reiserfs_journal_desc *)d_bh->b_data;
    commit = (struct reiserfs_journal_commit *)c_bh->b_data ;
    if (does_desc_match_commit (desc, commit)) {
	brelse (c_bh);
	return 1;
    }
    brelse (c_bh);
    return 0;
}


/* read the journal and find the oldest and newest transactions, return numbe
   of transactions fond */
int get_boundary_transactions (reiserfs_filsys_t * fs,
			       reiserfs_trans_t * oldest,
			       reiserfs_trans_t * newest)
{
    struct reiserfs_super_block * sb;
    unsigned long j_cur;
    unsigned long j_start;
    unsigned long j_size;
    struct buffer_head * d_bh;
    __u32 newest_trans_id, oldest_trans_id, trans_id;
    struct reiserfs_journal_desc * j_desc;
    int trans_nr;

    sb = fs->s_ondisk_sb;
    
    j_start = get_sb_journal_1st_block (fs->s_ondisk_sb);
    j_size = get_sb_journal_size (fs->s_ondisk_sb);
    
    oldest_trans_id = 0xffffffff;
    newest_trans_id = 0;

    trans_nr = 0;
    for (j_cur = 0; j_cur < j_size; j_cur ++) {
	d_bh = bread (fs->s_journal_dev, j_start + j_cur, fs->s_blocksize);

	if (!is_valid_transaction (fs, d_bh)) {
	    brelse (d_bh);
	    continue;
	}

	trans_nr ++;

	j_desc = (struct reiserfs_journal_desc *)d_bh->b_data;
	
	trans_id = get_jdesc_trans_id (j_desc);
	if (trans_id < oldest_trans_id) {
	    oldest_trans_id = trans_id;

	    oldest->mount_id = get_jdesc_mount_id (j_desc);
	    oldest->trans_id = get_jdesc_trans_id (j_desc);
	    oldest->desc_blocknr = d_bh->b_blocknr;
	    oldest->trans_len = get_jdesc_len (j_desc);
	    oldest->commit_blocknr = commit_expected (fs, d_bh);
	}

	if (trans_id > newest_trans_id) {
	    newest_trans_id = trans_id;

	    newest->mount_id = get_jdesc_mount_id (j_desc);
	    newest->trans_id = get_jdesc_trans_id (j_desc);
	    newest->desc_blocknr = d_bh->b_blocknr;
	    newest->trans_len = get_jdesc_len (j_desc);
	    newest->commit_blocknr = commit_expected (fs, d_bh);
	}

	j_cur += get_jdesc_len (j_desc) + 1;
	brelse (d_bh);
    }

    return trans_nr;
}


/* trans is a valid transaction. Look for valid transaction with smallest
   trans id which is bigger than the id of the current one */
int next_transaction (reiserfs_filsys_t * fs, reiserfs_trans_t * trans)
{
    struct buffer_head * d_bh, * next_d_bh;
    struct reiserfs_journal_desc * next_desc;
    int found;
    unsigned long j_start;
    unsigned long j_offset;
    unsigned long block;


    j_start = get_sb_journal_1st_block (fs->s_ondisk_sb);

    found = 0;

    /* make sure that 'trans' is a valid transaction */
    d_bh = bread (fs->s_journal_dev, trans->desc_blocknr, fs->s_blocksize);
    if (!is_valid_transaction (fs, d_bh))
	die ("next_transaction: valid transaction is expected");

    block = next_desc_expected (fs, d_bh);
    j_offset = block - j_start;

    while (1) {
	next_d_bh = bread (fs->s_journal_dev, block, fs->s_blocksize);
	if (is_valid_transaction (fs, next_d_bh))
	    break;

	brelse (next_d_bh);
	j_offset ++;
	block = j_start + (j_offset % get_sb_journal_size (fs->s_ondisk_sb));
    }

    next_desc = (struct reiserfs_journal_desc *)next_d_bh->b_data;
    
    if (trans->trans_id < get_jdesc_trans_id (next_desc)) {
	/* found transaction is newer */
	trans->mount_id = get_jdesc_mount_id (next_desc);
	trans->trans_id = get_jdesc_trans_id (next_desc);
	trans->desc_blocknr = next_d_bh->b_blocknr;
	trans->trans_len = get_jdesc_len (next_desc);
	trans->commit_blocknr = commit_expected (fs, next_d_bh);
	found = 1;
    }

    brelse (d_bh);
    brelse (next_d_bh);
    return found;
}


/* transaction is supposed to be valid */
int replay_one_transaction (reiserfs_filsys_t * fs,
			    reiserfs_trans_t * trans,
			    FILE * log_fp)
{
    struct buffer_head * d_bh, * c_bh;
    struct reiserfs_journal_desc * desc;
    struct reiserfs_journal_commit * commit;
    unsigned long j_start, j_offset, j_size;
    int i;
    

    d_bh = bread (fs->s_journal_dev, trans->desc_blocknr, fs->s_blocksize);
    if (!is_valid_transaction (fs, d_bh))
	die ("replay_one_transaction: not a valid transaction");

    c_bh = bread (d_bh->b_dev, commit_expected (fs, d_bh),
		  fs->s_blocksize);
    if (!c_bh)
	die ("replay_one_transaction: could not read commit block");

    desc = (struct reiserfs_journal_desc *)d_bh->b_data;
    commit = (struct reiserfs_journal_commit *)c_bh->b_data;

    /* first block of journal */
    j_start = get_sb_journal_1st_block (fs->s_ondisk_sb);
    j_size = get_sb_journal_size (fs->s_ondisk_sb);

    /* offset in the journal where the transaction starts */
    j_offset = trans->desc_blocknr - j_start + 1;


    print_one_transaction (log_fp, fs, trans);

    for (i = 0; i < trans->trans_len; i ++, j_offset ++) {
	struct buffer_head * in_place, * in_journal;
	unsigned long block;

	in_journal = bread (d_bh->b_dev, j_start + (j_offset % j_size), fs->s_blocksize);
	if (!in_journal) {
	    if (log_fp)
		fprintf (log_fp,
			 "replay_one_transaction: transaction %lu: "
			 "reading %lu block failed\n",
			 trans->trans_id, j_start + (j_offset % j_size));
	    continue;
	}

	if (i < JOURNAL_TRANS_HALF) {
	    block = le32_to_cpu (desc->j2_realblock[i]);
	} else {
	    block = le32_to_cpu (commit->j3_realblock[i - JOURNAL_TRANS_HALF]);
	}
	
	if (not_journalable (fs, block)) {
	    if (log_fp)
		fprintf (log_fp,
			 "replay_one_transaction: transaction %u: "
			 "block %d could not be replayed to (%lu)\n",
			 get_jdesc_trans_id (desc), i, block);
	} else {
	    in_place = getblk (fs->s_dev, block, fs->s_blocksize) ;
	    memcpy (in_place->b_data, in_journal->b_data, fs->s_blocksize);
	    mark_buffer_dirty (in_place);
	    mark_buffer_uptodate (in_place, 1);
	    bwrite (in_place);
	    brelse (in_place);
	}
	brelse (in_journal);
    }

    brelse (d_bh);
    brelse (c_bh);
    return 0;
}


void for_each_transaction (reiserfs_filsys_t * fs, actionf_t action)
{
    reiserfs_trans_t oldest, newest;


    if (!get_boundary_transactions (fs, &oldest, &newest))
	return;

    while (1) {
	action (fs, &oldest);
	if (!next_transaction (fs, &oldest))
	    break;
    }
}
